{-# LANGUAGE RankNTypes, KindSignatures, GADTs, MultiParamTypeClasses, FlexibleInstances, ScopedTypeVariables #-}
module Section6Encodings where

import Data.Function (fix)
import Data.List (intercalate)
import Section4TypedLists
import Section5TypedASGs (ttoList)

-- Encoding datatypes as type classes

class ArithAlg expr (f :: * -> *) where
  lit   :: Int -> expr f Int
  (.+.) :: expr f Int -> expr f Int -> expr f Int

-- Encoding Binders

class BindAlg expr f where
  var    :: f t -> expr f t
  letrec :: CList ts => (TList f ts -> TList (expr f) ts) ->
                        (TList f ts -> expr f t) -> expr f t

let_ :: BindAlg expr f => expr f t1 ->
                          (expr f t1 -> expr f t2) -> expr f t2
let_ e1 e2 = letrec (\ _ -> e1 ::: TNil) (\ (x ::: TNil) -> e2 (var x))

-- Generic behavior for binders

newtype Default f t = D { unD :: f t }

instance BindAlg Default f where
  var x       = D x
  letrec es e = e (fix (tmap unD . es))

-- Extensibility

class LamAlg expr f where
  lam :: (f t1 -> expr f t2) -> expr f (t1 -> t2)
  app :: expr f (t1 -> t2) -> expr f t1 -> expr f t2

-- Not shown in paper:
(.@.) :: LamAlg expr f => expr f (t1 -> t2) -> expr f t1 -> expr f t2
(.@.) = app

class (BindAlg expr f, ArithAlg expr f, LamAlg expr f) =>
      ExprAlg expr f

type ClosedExpr t = forall expr f. ExprAlg expr f => expr f t

-- Evaluation

newtype Eval t = E { unE :: t }

eval :: Default Eval t -> t
eval = unE . unD

toE :: t -> Default Eval t
toE = D . E

instance ArithAlg Default Eval where
  lit i     = toE i
  e1 .+. e2 = toE (eval e1 + eval e2)

instance LamAlg Default Eval where
  lam f     = toE (eval . f . E)
  app e1 e2 = toE ((eval e1) (eval e2))

instance ExprAlg Default Eval

-- Shared trees

one :: ArithAlg expr f => expr f Int
one = lit 1

tree_I 0 = one
tree_I n = let shared = tree_I (n - 1) in shared .+. shared

tree_E 0 = one
tree_E n = let_ (tree_E (n - 1)) (\ shared -> shared .+. shared)

test_tree_I_1 = eval (tree_I 2)
test_tree_I_2 = eval (tree_I 30) -- fast
test_tree_E_1 = eval (tree_E 2)
test_tree_E_2 = eval (tree_E 30) -- fast

-- Not shown in paper: definition of 'double'
newtype Dbl t = Dbl { unDbl :: ClosedExpr t }

double :: Default Dbl t -> ClosedExpr t
double = unDbl . unD

toDbl :: ClosedExpr t -> Default Dbl t
toDbl = D . Dbl

instance ArithAlg Default Dbl where
  lit i     = toDbl (lit (2 * i))
  e1 .+. e2 = toDbl (double e1 .+. double e2)

test_eval_double = eval (double (tree_I 30)) -- slow

-- Printing terms

newtype Text (f :: * -> *) t = T { text' :: Int -> String }

instance BindAlg Text (K String) where
  var x = T (\ _ -> unK x)
  letrec es e = T (\ c ->
    let vs = tmapK (\i -> "v" ++ show i) (tenumFrom c)
        c' = c + tlength vs
        ds = ttoList $
             tzipWith (\ (K v) e -> K (v ++ " = " ++ text' e c'))
                      vs (es vs)
    in  "(let { " ++ intercalate "; " ds ++
        " } in " ++ text' (e vs) c' ++ ")")

-- Cases not shown in paper:
instance ArithAlg Text (K String) where
  lit n     = T (\ _ -> show n)
  e1 .+. e2 = T (\ c ->
    "(" ++ text' e1 c ++ " + " ++ text' e2 c ++ ")")

instance LamAlg Text (K String) where
  lam e     = T (\ c ->
    let v = "v" ++ show c
    in  "(\\ " ++ v ++ " -> " ++ text' (e (K v)) (c + 1))
  app e1 e2 = T (\ c ->
    "(" ++ text' e1 c ++ " " ++ text' e2 c ++ ")")

instance ExprAlg Text (K String)

text :: forall t. ClosedExpr t -> String
text e = text' (e :: Text (K String) t) 0

test_text_tree_I = text (tree_I 3)
test_text_tree_E = text (tree_E 3)
